#include "glx_symbols.h"

#include <dlfcn.h>
#include <stdio.h>

//#include "riva_glx.h"
#include "ati_symbols.h"

/* This is where we keep per-card information, like
 * cached info on where the mapped video ram is
 * This will be allocated as riva_sinfo[numscreens]
 */
ATI_GLX_INFO *ati_sinfo=NULL;


static void (*EnableVTSwitch)(void *);

static int any_error;
static void *my_dlsym(void *handle, const char *name)
{
   void *tmp = dlsym(handle, name);
   char *error;
   if ((error = dlerror()) != 0) {
      fputs(error, stderr);
      any_error = 1;
      return 0;
   }
   return tmp;
}

/*
 *There are a few possible ways this nasty hack can be implemented:
 * The "clean" way, with xfree doing it all.. or hacky ways.
 * Unfortunately, there is a bug or two in xfree vidmem management.
 * Sooo... hack away.
 *
 * Currently, what it does is look for unmanaged video memory that
 * is free because the user under-specified VideoRam in XF86Config.
 * We need around 8 megabytes free!! But we dont check that here.
 *
 * return true (1) on okay, false (0) on fail
 * riva_sinfo[].privMemBase,privMemSize   will be set here on okay.
 */
static int atiMapPrivMem(ScrnInfoPtr scrninfoP){
	int screennum=scrninfoP->scrnIndex;
	unsigned int byteoffset;
	char *vidphysaddr;
	int totalmem,freememsize;
	char *mappedaddr;

	/* Take total RAM size, subtract what xfree is using, and
	 * thats what we have left to play with.
	 */
	byteoffset=scrninfoP->videoRam * 1024;
	totalmem=(ati_sinfo[screennum].xfree_atirec->VideoRAM)*1024;
	freememsize=totalmem-byteoffset;

	vidphysaddr=(char*)ati_sinfo[screennum].xfree_atirec->LinearBase;

	/* I "happen to know" that the solaris implementation of
	 xf86MapVidMem() does not do boundary checks against the
	 value in XF86Config. That makes life a lot easier for us here.
	 */
fprintf(stderr,"Mapping Video Memory \nXF86 SIze =%d \nHadrware Size= %d \n",ati_sinfo[screennum].xfree_atirec->VideoRAM,byteoffset/1024,totalmem);

fprintf(stderr,"Mapping Video Memory \nFramebuffer @ %p LinearBase @ %p  \nFreemem %d \nByteoffset=%d\ntotalmem=%d",vidphysaddr,ati_sinfo[screennum].xfree_atirec->LinearBase,freememsize,byteoffset,totalmem);
fprintf(stderr,"LinearSize = %d \nFBPitch %d \nBytesPerPixel=%d\n",ati_sinfo[screennum].xfree_atirec->LinearSize,ati_sinfo[screennum].xfree_atirec->FBPitch,ati_sinfo[screennum].xfree_atirec->FBBytesPerPixel);
fprintf(stderr,"pMemory = %p \npShadow %p \npMMIO=%p\n",ati_sinfo[screennum].xfree_atirec->pMemory,ati_sinfo[screennum].xfree_atirec->pShadow,ati_sinfo[screennum].xfree_atirec->pMMIO);
fprintf(stderr,"depth = %d \nbitsPerPixel %d \n",ati_sinfo[screennum].xfree_atirec->depth,ati_sinfo[screennum].xfree_atirec->bitsPerPixel);
	mappedaddr=xf86MapVidMem(screennum,VIDMEM_FRAMEBUFFER,
				 ((unsigned int)vidphysaddr)+byteoffset,
				 freememsize);
	if(mappedaddr==NULL){
		ErrorF("Could not mmap extra mem for GLX purposes");
		return 0;
	}
	fprintf(stderr,
	        "DEBUG: atiMapPrivMem: mapped offset 0x%x(0x%p) to mem 0x%x\n",
		byteoffset, vidphysaddr, mappedaddr);

	ati_sinfo[screennum].privMemBase=mappedaddr;
	ati_sinfo[screennum].privMemOffset=byteoffset;
	ati_sinfo[screennum].privMemSize=freememsize;

	return 1;
}

/*
 * Very important routine.
 * Not only is this routine responsible for "Hooking in XFree server symbols"
 * but also for a lot of large memory allocation:
 *    BACK BUFFER
 *    DEPTH BUFFER
 *    TEXTURE [memory, not exactly a "buffer"]
 * I would LIKE to do this in the more appropriately named atiInitGLX,
 * however, that routine depends on this stuff already being set up.
 *
 * return true (1) on okay, false (0) on fail
 */
int atiHookServerSymbols( void *handle )
{
   int screennum=0; /* XXX hardcode screennum. Bad Boy */
   ScrnInfoPtr scrninfoP;
   ScreenPtr screenP;
   ATIPtr ATIptr;
   int bytesperpixel, dbuffersize,depthbufsize, texturebufsize;
   char *currfreeptr;
   int freeoffset, freebytesleft;

   fprintf(stderr,"DEBUG: in atiHookServerSymbols\n");
   any_error = 0;


   /* Hook into symbols that are in the XFree nv driver specifically */
   if(ati_sinfo !=NULL){
	ErrorF("atiHookServerSymbols: riva_sinfo not NULL?\n");
	return 0;
   }
  scrninfoP=glxsym.scrninfoList[screennum];
   screenP=screenInfo.screens[scrninfoP->scrnIndex];

   ATIptr=ATIPTR ( scrninfoP);
   fprintf(stderr,"Number of screens= %d\n",screenInfo.numScreens);
  ati_sinfo=(ATI_GLX_INFO*)xalloc(sizeof (ATI_GLX_INFO) * screenInfo.numScreens);
   if(!ati_sinfo){
	ErrorF("atiHookServerSymbols: xalloc failed\n");
	return 0;
   }

  ati_sinfo[screennum].pScrnInfo=glxsym.scrninfoList[screennum];


   bytesperpixel=(scrninfoP->bitsPerPixel) >>3;

   if(scrninfoP->bitsPerPixel & 0xf){
	/* 15 bits = 1 + 1 bytes*/
	bytesperpixel+=1;
   }

   ati_sinfo[screennum].xfree_atirec= ATIptr;
   ati_sinfo[screennum].atihw=&ATIptr->NewHW;
fprintf(stderr,"LinearBase @ %p totalmem=%d",ati_sinfo[screennum].xfree_atirec->LinearBase,ATIptr->VideoRAM);
fprintf(stderr,"LinearSize = %d \nFBPitch %d \nBytesPerPixel=%d\n",ati_sinfo[screennum].xfree_atirec->LinearSize,ati_sinfo[screennum].xfree_atirec->FBPitch,ati_sinfo[screennum].xfree_atirec->FBBytesPerPixel);
fprintf(stderr,"pMemory = %p \npShadow %p \npMMIO=%p\n",ati_sinfo[screennum].xfree_atirec->pMemory,ati_sinfo[screennum].xfree_atirec->pShadow,ati_sinfo[screennum].xfree_atirec->pMMIO);
fprintf(stderr,"depth = %d \nbitsPerPixel %d \n",ati_sinfo[screennum].xfree_atirec->depth,ati_sinfo[screennum].xfree_atirec->bitsPerPixel);
fprintf(stderr,"Attempting to map Buffers\n");
{
int ii;
for(ii=0;ii<sizeof(*ATIptr);ii++){
	fprintf(stderr,"%02X ",*((char *)ATIptr+ii));
	if((ii%40) == 0) fprintf(stderr,"\n");
}
}

/*
Just for the moment bypass all the buffer allocation stuff
*/
goto bypass;

   if(!atiMapPrivMem(scrninfoP)){
	ErrorF("atiHookServerSymbols bailing out\n");
	return 0;
   }


   currfreeptr=ati_sinfo[screennum].privMemBase;
   freebytesleft=ati_sinfo[screennum].privMemSize;
   freeoffset=ati_sinfo[screennum].privMemOffset; /*offset from vidram 0*/

   ati_sinfo[screennum].videoareasize=bytesperpixel*
				scrninfoP->virtualX * scrninfoP->virtualY;
   dbuffersize=ati_sinfo[screennum].videoareasize;

   /************************************************************************
    * Now starts all the big backbuffer, etc...  allocation.               *
    *                                                                      *
    * We are more likely to need a "back buffer" for doublebuffering,      *
    * then a "depth buffer". But try for both.                             *
    * Due to nastiness in xfree memory management routines, we have done   *
    * our own reserving of some videoram space. We now likewise do our own *
    * "memory allocation" from a superchunk that we have hanging off       *
    * currfreeptr. SIGH.
    ***********************************************************************/

   /* This is just assumed to always be 0, but it's a consistency thing
    * to have it defined .
    */
   ati_sinfo[screennum].frontbufferoffset=0;


   /* Allocate "Depth Buffer", for z-buffering */
   /* It seems the TNT hardware is always 16bits for depth buffers.
    * See RivaGetDepth(). If this changes, will have to
    * initialize better here
    */
   ati_sinfo[screennum].depth_bits=16;

   depthbufsize=(ati_sinfo[screennum].depth_bits>>3) *
		scrninfoP->virtualX * scrninfoP->virtualY;

   if(freebytesleft<depthbufsize){
	ErrorF("atiHookServerSymbols: alloc of depth buffer failed: %d free, need %d\n",
	       freebytesleft, depthbufsize);
	return 0;
   }

   ati_sinfo[screennum].depthbufferoffset=freeoffset;
   ati_sinfo[screennum].depthbufmappedaddr=currfreeptr;
   freeoffset +=depthbufsize; currfreeptr +=depthbufsize;
   freebytesleft -=depthbufsize;


   /*
    * Allocate "Back Buffer", for double--buffering
    */
   if(freebytesleft<dbuffersize){
	ErrorF("atiHookServerSymbols: alloc of back buffer failed: %d free, need %d\n",
	       freebytesleft, dbuffersize);
	return 0;
   }

   ati_sinfo[screennum].backbufferoffset=freeoffset;
   ati_sinfo[screennum].backbufmappedaddr=currfreeptr;
   freeoffset +=dbuffersize; currfreeptr +=dbuffersize;
   freebytesleft -=dbuffersize;


   /* Allocate "Texture (Area)", for loading textures onto the card */
   /* XXX in future, calculate numtexblocks better, based on amount of vidmem*/
   /* Heck, in the future, we should be using AGP instead, somehow. */
   ati_sinfo[screennum].numtexblocks=ATIGLX_NUMTEXTUREBLOCKS;

   /* XXX we could actually do this the other way around, and
    * set numtexblocks based on amount of free mem.
    * But not right now.
    */
   texturebufsize=ati_sinfo[screennum].numtexblocks *ATI_TEX_BLOCK_SIZE;

   if(freebytesleft<texturebufsize){
	ErrorF("atiHookServerSymbols: alloc of texture buffer failed: %d free, need %d\n",
	       freebytesleft, texturebufsize);
	return 0;
   } else
   {
	int loop;
	for(loop=0; loop<ati_sinfo[screennum].numtexblocks; loop++){
		ati_sinfo[screennum].textureoffset[loop]=
			freeoffset + (loop*ATI_TEX_BLOCK_SIZE);
		ati_sinfo[screennum].texturebase[loop]=
			currfreeptr + (loop*ATI_TEX_BLOCK_SIZE);
	}
	/* Note that we do anything else with this, but
	 * lets be consistent
	 */
	freeoffset +=texturebufsize; currfreeptr +=texturebufsize;
	freebytesleft -=texturebufsize;
   }
// Nasty hack to bypass the buffer allocations
bypass:
   ErrorF("   GLX disabling VT switching\n");

   EnableVTSwitch=my_dlsym(handle,"xf86EnableVTSwitch");

   if(EnableVTSwitch){
	ErrorF("Disabling VT switching, for initial GLX version\n");
	(*EnableVTSwitch)(FALSE);
   }

   if (!any_error)
      ErrorF("Sucessfully linked in GLX ATI subdriver\n");
   else
      ErrorF("GLX ATI driver disabled due to missing symbols\n");

   fprintf(stderr,"DEBUG: atiHook wrapup: depthoffset=%d, backoffset=%d, texture=%p\n",
		ati_sinfo[0].depthbufferoffset,
		ati_sinfo[0].backbufferoffset,
		ati_sinfo[0].textureoffset[0]);

   return !any_error;
}
