/* This file is badly named. It should be called glx_mod.c or something */

/*
 * GLX Server Extension
 * Copyright (C) 1996  Steven G. Parker  (sparker@cs.utah.edu)
 * Copyright (C) 1998, 1999 Terence Ripperda (ripperda@sgi.com)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * STEVEN PARKER, TERENCE RIPPERDA, OR ANY OTHER CONTRIBUTORS BE LIABLE FOR
 * ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/* TODO:
 - destroy context should not destroy until the context is made uncurrent
 - pixmaps
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <assert.h>
#include <signal.h>

#include <dlfcn.h>

#include "extnsionst.h"
#include "dixstruct.h"
#include "scrnintstr.h"
#include "pixmapstr.h"
#include "windowstr.h"
#include "GL/GLXproto.h"
#include "xsmesaP.h"
#include "xsmesa.h"
#include "config.h"
#include "glx_log.h"
#include "glx_config.h"
#include "glxcmds.h"
#include "glxrender.h"
#include "glxdecode.h"
#include "fontstruct.h"
#include "glxvisual.h"

#include "glx_macros.h"
#include "glx_clients.h"
#include "glx_dispatch.h"
#include "glx_symbols.h"
#include "xf86Version.h"
#include "GL/xmesa.h"

#ifdef HAVE_FLOATINGPOINT_H
#include <floatingpoint.h>
#endif

RESTYPE glContexts;
RESTYPE glPixmaps;
RESTYPE glWindows;
RESTYPE glxClients;

#include "xf86Module.h"

/* #include <vga.h> */

#include "../hwglx/common/glx_init.h"
#include "../hwglx/common/glx_symbols.h"
#include "common/xf86.h"

unsigned int glx_chipset = SOFTWARE_MODE;
unsigned int glx_server = XF_UNKNOWN;

extern XSMesaVisual GLfind_vis(VisualID vid);

GLX_PROCS GLXProcs;

static int ServerMesaDispatch(register ClientPtr client);
static int ServerMesaDispatch_swapped(register ClientPtr client);

static void ServerMesaReset(ExtensionEntry* extEntry)
{
    /* Ideally, this function should free all memory allocated by GLX and
     * Mesa. Right now we just do enough so we can avoid a server crash */

    ErrorF("%s Extension Reset called!\n", GLX_EXTENSION_NAME);

    log_finish();
    glx_freeconfig();

    if(GLXProcs.ResetDriver){
      GLXProcs.ResetDriver();
    }

}

int GLDestroyMesaContext(pointer value, XID id)
{
    XSMesaContext ctx;
    fprintf(stderr, "destroying a context\n");
    ctx=(XSMesaContext)value;
    log_print(LOG_ALWAYS, "Destroying context\n");
    GLXProcs.DestroyContext(ctx);
    return 0;
}

int GLDestroyMesaBuffer(pointer value, XID id)
{
    XSMesaBuffer buf=(XSMesaBuffer)value;
    fprintf(stderr, "destroying a buffer\n");
    log_print(LOG_ALWAYS, "Destroying buffer\n");
    GLXProcs.DestroyBuffer(buf);
    return 0;
}

int GLDestroyGLXClients(pointer value, XID id)
{

    fprintf(stderr, "destroying a client: %u!\n", (unsigned int) id);
    log_print(LOG_ALWAYS, "destroy client\n");

    __glXRemoveClient((int)value);
    /* xfree(client); */
    return 0;
}

static int GLNoDriverPrivates( ClientPtr client,
			       XSMesaContext ctx,
			       xGLXVendorPrivateReq *stuff )
{
   ErrorF("GLNoDriverPrivates");
   return GLXUnsupportedPrivateRequest;
}


static int GLDenyDirect( ClientPtr client )
{
   return 0;
}

static void GLFrontBufferAlwaysUpToDate( DrawablePtr b )
{
}


static void (*OldSignalHandler)(int);

void
glxSignalHandler(int sig)
{
   /* clean up any logging before dying */
   log_print(LOG_ALWAYS, "GLX module Caught signal %d\n", sig);
   log_finish();
   OldSignalHandler(sig);
}

/* I am not sure this old routine is valid for use with xfree4.0
 * Guess we will find out
 */
void
glxInitLogs() {
   fprintf(stderr, "Initializing logs\n");

   /* set up logging capabilities */
   if (set_log_file("/var/log/glx_debug.log")) {
      fprintf(stderr, "glX Error: error setting log file\n");
   }
   log_initialize();
   log_print(LOG_ALWAYS, "OpenGL/GLX Extension log file\n");

   /* Install signal handler */
   /* Which signals else do we need?  -Wittawat */
   /* stolen from mga/g200 driver */
   OldSignalHandler = signal(SIGSEGV,glxSignalHandler);
   if (OldSignalHandler == SIG_ERR) {
      log_print(LOG_ALWAYS, "Could not install signal handler!\n");
      ErrorF("GLX: couldn't install signal handler!\n");
   }

   /* log some configuration information */
   log_print(LOG_ALWAYS, "Configuration info:\n");
#ifdef MESA31
   log_print(LOG_ALWAYS, "Mesa: 3.1 v.%d\n", MESA_VERSION);
#else
   log_print(LOG_ALWAYS, "Mesa: 3.0\n");
#endif
   {
#ifdef MGA
      log_print(LOG_ALWAYS, "Hardware accelerated: mga (Matrox G200/G400)\n");
#endif
#ifdef TNT
      log_print(LOG_ALWAYS, "Hardware accelerated: tnt (NVidia Riva series)\n");
#endif
#ifdef I810
      log_print(LOG_ALWAYS, "Hardware accelerated: i810\n");
#endif
#ifdef S3VIRGE
      log_print(LOG_ALWAYS, "Hardware accelerated: S3 ViRGE\n");
#endif
#ifdef S3SAVAGE
      log_print(LOG_ALWAYS, "Hardware accelerated: S3 Savage3D\n");
#endif
#ifdef SIS6326
      log_print(LOG_ALWAYS, "Hardware accelerated: sis6326\n");
#endif
#ifdef MACH64
      log_print(LOG_ALWAYS, "Hardware accelerated: mach64 (ATI Rage Pro)\n");
#endif
      log_print(LOG_ALWAYS, "All other videocards fallback to software\n");
   }
   if(glx_getint("nohw")){
      log_print(LOG_ALWAYS, "Forced to software mode\n");
   } 
#ifdef USE_X86_ASM
   log_print(LOG_ALWAYS, "Compiled with x86 assembly support\n");
#endif
#ifdef USE_MMX_ASM
   log_print(LOG_ALWAYS, "Compiled with MMX assembly support\n");
#endif
#ifdef USE_3DNOW_ASM
   log_print(LOG_ALWAYS, "Compiled with 3dNow! assembly support\n");
#endif
   log_print(LOG_ALWAYS, "End Configuration info\n");
   log_flush();
}


/* This is the actual "Initialize your module, we're going live" function.
 * Note that this does wierd things to open "ourself", so that
 * we can dlsym() locate extra functions that are hardware-specific.
 * WILL CLEAN THIS UP LATER. But for right now, should probably work
 * fine. So let it ride.
 *
 * Calls:
 *	glxInitLogs()
 *	CreateNewResourceType([various contexts, blah blah])
 *	AddExtention() [xfree critical hook]
 * Initializes GLXProcs.blahblahblah - defaults to software renderer, then
 *	lets hardware entry points like mgaHookServerSymbols() update
 *	things for hardware accel.
 * calls xxxInitGLX()  [example, mgaInitGLX]
 *	GLFinalizeVisuals(), __glXInitClients()
 */
void GlxExtensionInit()
{
   ExtensionEntry* extEntry;
   void *handle;
   ScreenPtr pScreen;

   int screennum=0;
	/* XXX oops. hardcoding to screen 0. am I a bad boy? */
	/* We should actually go through and have separate inits
	 * for each screen!!
	 * For now, use 'screennum' rather than actual 0 lower down.
	 */

   pScreen=screenInfo.screens[screennum];

   fprintf(stderr, "DEBUG: GlxExtensionInit called\n");
#ifdef __FreeBSD__
   fpsetmask(0); /* put the FPU in IEEE mode (mask all exceptions) */
#endif

   handle = dlopen (0, RTLD_LAZY);
   if (!handle) {
      fputs (dlerror(), stderr);
      fprintf(stderr, "Cannot self-dlopen - GLX module disabled\n");
   }

   if( glx_readconf(GLXCONF) == CONFIG_FATAL) {
      fprintf(stderr, "An error occured while parsing %s - GLX disabled\n", 
	      GLXCONF);
      return;
   }

   /* It is only safe to use standard xfree functions like ErrorF,
    * AFTER our symbol table is filled out!
    */
   if(!glxHookXFree86Symbols( handle )){
	fprintf(stderr,"GLX could not resolve X server symbols. Unloading\n");
	dlclose(handle);
	return; /* will magically auto-unload somehow */
   }

/*
 * These blow things up, unless the special stupid "stub"
 * is compiled into the X-server.
 * I think that is ugly, so am going to try to code around it.
 *
   GlxExtensionInitPtr = GlxExtensionInit;
   GlxInitVisualsPtr   = GlxInitVisuals;
*/

   glxInitLogs();
fprintf(stderr,"Initialised Logs\n");
#if defined(MESA_MAJOR_VERSION) && defined(MESA_MINOR_VERSION)
    ErrorF(
       "\t%s extension module for XFree86 %d.%d.%d-- Mesa version %d.%d\n\t\tGLX package version %s, GLX protocol version %s.\n",
       GLX_EXTENSION_NAME,
       XF86_VERSION_MAJOR,XF86_VERSION_MINOR,XF86_VERSION_PATCH,
       MESA_MAJOR_VERSION, XMESA_MINOR_VERSION,
       GLX_PACKAGE_VERSION, GLX_PROTOCOL_VERSION
    );
#else
    ErrorF(
       "\t%s extension module for XFree86%d.%d.%d--\n\t\tGLX package version %s, GLX protocol version %s.\n",
       GLX_EXTENSION_NAME,
       XF86_VERSION_MAJOR,XF86_VERSION_MINOR,XF86_VERSION_PATCH,
       GLX_PACKAGE_VERSION, GLX_PROTOCOL_VERSION
    );
#endif
fprintf(stderr,"Creating Resources\n");
   glContexts = CreateNewResourceType(GLDestroyMesaContext);
   glWindows  = CreateNewResourceType(GLDestroyMesaBuffer);
   glPixmaps  = CreateNewResourceType(GLDestroyMesaBuffer);
   glxClients = CreateNewResourceType(GLDestroyGLXClients);
fprintf(stderr,"Adding Extension to Server\n");
   extEntry = AddExtension(GLX_EXTENSION_NAME, GLX_NUMBER_EVENTS,
			   GLX_NUMBER_ERRORS, ServerMesaDispatch,
			   ServerMesaDispatch_swapped, ServerMesaReset,
			   StandardMinorOpcode);

   if( !extEntry ) {
      ErrorF("GlxExtensionInit(): AddExtension() failed\n" );
      dlclose(handle);
      return;
   }

   __glxErrorBase = extEntry->errorBase;
   __glxReqBase   = extEntry->base;
   __glxEventBase = extEntry->eventBase;

   /*
    * Set default XSMesa procs.
    */
   GLXProcs.CreateVisual       = XSMesaCreateVisual;
   GLXProcs.DestroyVisual      = XSMesaDestroyVisual;
   GLXProcs.CreateContext      = XSMesaCreateContext;
   GLXProcs.DestroyContext     = XSMesaDestroyContext;
   GLXProcs.MakeCurrent        = XSMesaMakeCurrent;
   GLXProcs.CopyContext        = XSMesaCopyContext;
   GLXProcs.CreateWindowBuffer = XSMesaCreateWindowBuffer;
   GLXProcs.CreatePixmapBuffer = XSMesaCreatePixmapBuffer;
   GLXProcs.DestroyBuffer      = XSMesaDestroyBuffer;
   GLXProcs.BindBuffer         = XSMesaBindBuffer;
   GLXProcs.FindBuffer         = XSMesaFindBuffer;
   GLXProcs.SwapBuffers        = XSMesaSwapBuffers;
   GLXProcs.GetBackBuffer      = XSMesaGetBackBuffer;
   GLXProcs.GetDepthBuffer     = XSMesaGetDepthBuffer;
   GLXProcs.GetCurrentContext  = XSMesaGetCurrentContext;
   GLXProcs.GetCurrentBuffer   = XSMesaGetCurrentBuffer;
   GLXProcs.FeedbackBuffer     = XSMesaFeedbackBuffer;
   GLXProcs.SelectionBuffer    = XSMesaSelectionBuffer;
   GLXProcs.GetPixel           = GLXGetPixel;
   GLXProcs.PutPixel           = GLXPutPixel;
   GLXProcs.DestroyImage       = GLXDestroyImage;
   GLXProcs.CreateImage        = GLXCreateImage;
   GLXProcs.GetDepth           = GLXGetDepth;
   GLXProcs.PutDepth           = GLXPutDepth;
   GLXProcs.CreateDepthBuffer  = GLXCreateDepthBuffer;
   GLXProcs.VendorPrivate      = GLNoDriverPrivates;
   GLXProcs.AllowDirect        = GLDenyDirect;
   GLXProcs.ValidateFrontBuffer = GLFrontBufferAlwaysUpToDate;


   /* Runtime detect which server has loaded the glx.so module.
    */
   if (!glx_getint("nohw")) {
const char *chipset;
      /* Don't want to do this on the client side:
       */
      if (__glx_is_server) {
	 glx_server = XF_SVGA;
	 glx_chipset = SOFTWARE_MODE;

	 if (glxHookSVGAServerSymbols( handle )) {
	    chipset = glxsym.scrninfoList[screennum]->chipset;

	    log_print(LOG_TRACE,
		      "Detected XF86_SVGA server, "
		      "looking for chipset '%s'\n",
		      chipset);
	    glx_server = XF_SVGA;

	    if (0)
	    	;
#ifdef MGA
	    else if (!strcmp("mgag400", chipset) &&
		     mgaHookServerSymbols( handle ))
	    {
	       glx_chipset = MGA_G400;
	       log_print(LOG_ALWAYS, "Using G400 driver!\n");
	    }
	    else if (!strcmp("mgag200", chipset) &&
		     mgaHookServerSymbols( handle )) {
	       glx_chipset = MGA_G200;
	       log_print(LOG_ALWAYS, "Using G200 driver!\n");
	    }
#endif
#ifdef TNT
	    else if ((!strcmp("RIVA128",  chipset) ||
		      !strcmp("RIVATNT",  chipset) ||
		      !strcmp("RIVATNT2", chipset) ||
		      !strcmp("RIVA 128", chipset) ||
		      !strcmp("RIVA TNT", chipset) ||
		      !strcmp("RIVA TNT2", chipset) ||
		      !strcmp("RIVA ULTRA TNT2", chipset) ||
		      !strcmp("RIVA TNT2 Ultra", chipset) ||
		      !strcmp("RIVA VANTA", chipset) ||
		      !strcmp("RIVA ULTRA VANTA", chipset) ||
		      !strcmp("RIVA INTEGRATED", chipset) ||
		      !strcmp("GeForce 256", chipset) ||
		      !strcmp("GeForce DDR", chipset) ||
		      !strcmp("Quadro", chipset)))
	   {
	     log_print(LOG_ALWAYS,"TNT chipset matched...\n");
		if(nvHookServerSymbols( handle )) {
			glx_chipset = NV_RIVATNT;
			log_print(LOG_ALWAYS,
			 "Using TNT driver (no direct rendering)\n");
		} else {
			log_print(LOG_ALWAYS,
			 "nvHookServerSymbols FAILED\n");
		}
	   }

#endif
#ifdef S3VIRGE
	    else if ( !strcmp("s3_virge", chipset) &&
		      s3virgeHookServerSymbols( handle )) {
	       glx_chipset = S3_VIRGE;
	       log_print(LOG_ALWAYS,
			 "Using S3 ViRGE driver (no direct rendering)!\n");
	    }
#endif
#ifdef S3SAVAGE
            else if (strcmp("s3_savage", chipset) == 0
                      && s3savHookServerSymbols(handle))
                    {
                       glx_chipset = S3_SAVAGE3D;
                       log_print(LOG_ALWAYS, "Using Savage3D driver\n");
                    }
#endif
#ifdef SIS6326
	    else if ( !strcmp("sis6326", chipset) &&
		      sis6326HookServerSymbols( handle )) {
	       glx_chipset = SIS_6326;
	       log_print(LOG_ALWAYS,
			 "Using SiS 6326 driver (no direct rendering)!\n");
	    }
#endif
#ifdef I810
	    else if ( (!strcmp("i810", chipset) ||
		       !strcmp("i810e", chipset) ||
		       !strcmp("i810-dc100", chipset)) &&
		      i810HookServerSymbols( handle )) {
	       glx_chipset = I_810;
	       log_print(LOG_ALWAYS, "Using Intel i810 driver!\n");
	    }
#endif
 // }
#ifdef MACH64
	    else if (!strcmp("ati", chipset))
	    {
	 	log_print(LOG_ALWAYS,"Attempt attach to RagePRO driver\n");
		if (glxHookMach64ServerSymbols( handle )) {
	    		log_print(LOG_ALWAYS, "Using RagePRO driver under XF86_Mach64\n");
	    		glx_chipset = ATI_RAGEPRO;
	   	 	glx_server = XF_MACH64;
	 	}
	 }
#ifdef XFREE3
/*
fbDev server not supprted in XF4
*/


	 else if (glxHookFbDevServerSymbols( handle ) &&
		  strstr( GLXSYM(fbdevInfoRec).chipset, "Mach64" )) {
	    log_print(LOG_ALWAYS, "Using RagePRO driver under XF86_FbDev\n");
	    glx_chipset = ATI_RAGEPRO;
	    glx_server = XF_FBDEV;
	 }
#endif /*XFREE3*/ 
#endif
      }
  }
      /* Want to do this on both server and client:
       */
      switch (glx_chipset) {
#ifdef MGA
      case MGA_G200:
      case MGA_G400:
	 mgaInitGLX();
	 break;
#endif

#ifdef I810	
      case I_810:
	 i810InitGLX();
	 break;
#endif	 

#ifdef TNT
      case NV_RIVATNT:
	 hwInitVisuals = nvInitVisuals;
	 nvInitGLX();
	 break;
#endif

#ifdef MACH64
      case ATI_RAGEPRO:
	 mach64InitGLX();
	 break;
#endif

#ifdef S3VIRGE
      case S3_VIRGE:
	 s3virgeInitGLX();
	 break;
#endif

#ifdef S3SAVAGE
      case S3_SAVAGE3D:
	s3savInitGLX();
	break;
#endif

#ifdef SIS6326
      case SIS_6326:
	 sis6326InitGLX();
	 break;
#endif

      case SOFTWARE_MODE:
	 log_print(LOG_ALWAYS, 
		   "Didn't find supported hardware, "
		   "rendering in software mode (slow)!!\n");
	 break;
      }
   }


   log_flush();

   /* This is  inspired from an earlier version of utah-glx,
    * from direct_protocol.c:__glx_init_direct().
    * Hope it works!
    */
   {
	int numvis,numdep,rootdep;
	VisualID rootvis;

	fprintf(stderr,"DEBUG: GlxExtensionInit: pscreen->numvis=%d\n",
		pScreen->numVisuals);

	numvis = pScreen->numVisuals; /* was '1' */
	numdep = pScreen->numDepths;
	rootdep = pScreen->rootDepth;
	rootvis = pScreen->rootVisual;

	/* Kinda funky. This adds new visuals and stuff to the arrays.
	 * But for some obscure reason, we cant pass in the pointer
	 * to the int counters, so pass back values by hand. for now.
	 */
      GlxInitVisuals(	&pScreen->visuals,
			&pScreen->allowedDepths,
			&numvis,
			&numdep,
			&rootdep,
			&rootvis,
			0,	/* sizes -- what is it? not used... */
			0 );	/* bitsperRGB -- in what visual? not used... */

	pScreen->numVisuals=numvis;
	pScreen->numDepths=numdep;
	/*pScreen->rootVisual=rootvis;*/

	fprintf(stderr,"DEBUG: GlxInitVisuals returned okay\n");
   }


   
   GLFinalizeVisuals();
   __glXInitClients();

	fprintf(stderr,"DEBUG: GlxExtensionInit done\n");

/* HACK: work around broken FreeBSD ld? courtesy of Simon Walton */
#ifndef __FreeBSD__
   dlclose(handle);
#endif
}

static int ServerMesaDispatch(register ClientPtr client)
{
    __glxClient *clientPtr = NULL;
    __glx_dispatch_proc dispatch_proc = NULL;
    int status = 0;
    REQUEST(xReq);

    clientPtr = __glXFindClient(client->index);
    if (clientPtr == NULL) {
       clientPtr = __glXAddClient(client);
       if (clientPtr == NULL) return BadAlloc;
    }


    log_print(LOG_TRACE, "glx op: %@\n", stuff->data);

    /* No need for a sanity check here, __glx_dispatch_table has 256 entries */
    dispatch_proc = __glx_dispatch_table[stuff->data];
    status = (*dispatch_proc)(client);

    log_print(LOG_TRACE, "glxop %@ done.\n", stuff->data);

    return status;
}

static int ServerMesaDispatch_swapped(register ClientPtr client)
{
    __glxClient *clientPtr = NULL;
    __glx_dispatch_proc dispatch_proc = NULL;
    int status = 0;
    REQUEST(xReq);

    clientPtr = __glXFindClient(client->index);
    if (clientPtr == NULL) {
       clientPtr = __glXAddClient(client);
       if (clientPtr == NULL) return BadAlloc;
    }


    log_print(LOG_TRACE, "glx op: %@\n", stuff->data);

    /** Currently, only Render and RenderLarge have _swapped versions **/
    /** Eventually, all of them should have duplicate versions **/


    if (stuff->data < 3) {
       if (stuff->data == 1) {
          return GLRender_swapped(client);
       }
       if (stuff->data == 2) {
          return GLRenderLarge_swapped(client);
       }
       return GLnoop(client);
    }

    /* No need for a sanity check here, __glx_dispatch_table has 256 entries */
    dispatch_proc = __glx_dispatch_table[stuff->data];
    status = (*dispatch_proc)(client);

    log_print(LOG_TRACE, "glxop %@ done.\n", stuff->data);

    return status;
}

int GLXDecodeInvalid(int length, char* data)
{
    ErrorF("2. INCOMPLETE: %d, length=%d\n", *(short*)(data-2), length);
    return Success;
}


extern Bool GlxInitVisuals(
    VisualPtr *         /*visualp*/,
    DepthPtr *          /*depthp*/,
    int *               /*nvisualp*/,
    int *               /*ndepthp*/,
    int *               /*rootDepthp*/,
    VisualID *          /*defaultVisp*/,
    unsigned long       /*sizes*/,
    int                 /*bitsPerRGB*/
);



#include "xf86.h"

#include "xf86Version.h"

ExtensionModule glxExtModinfo = {
    GlxExtensionInit,
    GLX_EXTENSION_NAME,
    NULL,NULL,NULL
    };

static XF86ModuleVersionInfo VersRec =
{
	"glx",
	"Utah-GLX",
	MODINFOSTRING1,
	MODINFOSTRING2,
	XF86_VERSION_CURRENT,
	1,0,0,
	ABI_CLASS_EXTENSION,
	ABI_EXTENSION_VERSION,
	MOD_CLASS_NONE,
	{0,0,0,0}
};

/* Not entirely sure what we should be returning here... 
 * But anyway, this seems to be the
 * "I'm loading your dymaic library into memory now" hook.
 * Not to be confused with the "Im INITIALIZING you now" hook, which is
 * GlxExtensionInit()
 */
static pointer
glxModuleSetup(pointer module, pointer opts, int *errmajor,int *errminor)
{
	static Bool setupdone=FALSE;

	if(setupdone==TRUE){
		if(errmajor){
			*errmajor=LDR_ONCEONLY;
		}
		return 0;
	}

	setupdone=TRUE;

	/* This will trigger GlxExtensionInit() */
	LoadExtension(&glxExtModinfo, FALSE);

	return (pointer)1;/* non-null means success */
}


XF86ModuleData glxModuleData = 
{
	&VersRec,
	glxModuleSetup, NULL
};
