
/*
 * GLX Server Extension
 * Copyright (C) 1996  Steven G. Parker  (sparker@cs.utah.edu)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * STEVEN PARKER, TERENCE RIPPERDA, OR ANY OTHER CONTRIBUTORS BE LIABLE FOR 
 * ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
 * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <stdlib.h>

#include "X.h"
#include "Xproto.h"
#include "scrnintstr.h"

#include "types.h"

#include "glxvisual.h"
#include "xsmesaP.h"
#include "glx_log.h"
#include <stdio.h>

#include "os.h"

#ifdef XFree86LOADER
#include "xf86_libc.h"
#endif

#include "../hwglx/common/glx_symbols.h"

GLScreen __glScreens[MAXSCREENS];

/* THis is a TEMPORARY list holder, to hold values between calling
 * GlxInitVisuals(), and GLFinalizeVisuals()
 */
Vis* __vislist;



/* Radically tweaked, in hopes of hacking something to actually WORK, dammit.
 * This creates a new virtual linked list of visuals, under __vislist.
 * This version is MAJORLY SIMPLIFIED from prior Utah-glx versions.
 * I'm trying to get the damn thing to WORK, not support individually
 * twiddling options like alpha, etc.
 *
 * Caller will later call GLFinalizeVisuals, which
 * then goes through that list, and makes Mesa-specific versions of the info.
 *
 * We do this in two steps, to allow for the possibility of a
 * hardware-specific InitVisuals call. See GlxInitVisuals().
 *
 * Pass in pointer to list of 'real' Visuals, along with associated values,
 * and we'll deref, but NOT change anything, at this point.
 * We just populate __vislist with our own allocated copies.
 */
Bool
XSMesaInitVisuals (visualp, depthp, nvisualp, ndepthp, rootDepthp, defaultVisp, sizes, bitsPerRGB)
VisualPtr   *visualp;
DepthPtr    *depthp;
int     *nvisualp, *ndepthp;
int     *rootDepthp;
VisualID    *defaultVisp;
unsigned long   sizes;
int     bitsPerRGB;
{
    int i;
    int nvisual=*nvisualp;

    int nongl_visuals=0;
    int gl_visuals=0;


    Vis* vhead=0;
    Vis* vtail=0;
    Vis* v;

	fprintf(stderr,"DEBUG: XSMesaInitVisuals: numvis=%d\n",nvisual);
	fprintf(stderr,"   DEBUG: defaultVisual has id=%ld\n",*defaultVisp);

    /* Go through all the visuals we have passed to us,
     * and see which ones are suitable for us to use.
     * Keep track of the "good" ones, on __vislist
     */
    for (i=0;i<nvisual;i++)
    {
        VisualPtr pVisual=(*visualp)+i;
        int ci, rgb;

	if (!pVisual->vid) {
	   fprintf(stderr, "ERROR: visual %d (out of %d) has zero vid\n",
		   i, nvisual);
	   abort();
	}

        /* See if OpenGL supports this visual... */
        ci=rgb=GL_FALSE;
        switch (pVisual->CLASS)
        {
            case StaticGray:
            case GrayScale:
                /* Any StaticGray/GrayScale visual works in RGB or CI mode */
                ci=GL_TRUE;
                rgb=GL_TRUE;
                break;
            case StaticColor:
            case PseudoColor:
                /* Any StaticColor/PseudoColor visual of at least 4 bits */
                if (pVisual->nplanes)
                {
                    ci=GL_TRUE;
                }
                else
                {
                    ci=GL_FALSE;
                }
                break;
            case TrueColor:
            case DirectColor:
                /* Any depth of TrueColor or DirectColor works in RGB mode */
                rgb=GL_TRUE;
                break;
            default:
                /* This should never happen */
                ci=GL_FALSE;
                rgb=GL_FALSE;
                break;
        }

	/* we like it? insert reference into __vlist */
        if (rgb)
        {
            int db, alpha, stencil, depth, accum;

	    fprintf(stderr,"DEBUG: adding RGB visual %ld\n",pVisual->vid);

            db=1;
            {
                alpha=1;
                {
                    stencil=1;
                    {
                        depth=1;
                        {
                            accum=1;
                            {
                                Vis* v=(Vis*)xalloc(sizeof(Vis));
                                v->orig_visual=pVisual;
				v->vid=pVisual->vid;
                                v->usegl=GL_TRUE;
                                v->rgb=GL_TRUE;
                                v->db=db;
                                v->alpha=alpha;
                                v->stencil=stencil;
                                v->depthb=depth;
                                v->accum=accum;
                                v->next=0;
                                if (vtail)
                                    vtail->next=v;
                                else
                                    vhead=v;
                                vtail=v;

                                gl_visuals++;
                            }
                        }
                    }
                }
            }
        }
        if (ci)
        {
            int db, stencil, depth;
	    fprintf(stderr,"DEBUG: adding CI visual %ld\n",pVisual->vid);
            db=1;
            {
                stencil=1;
                {
                    depth=1;
                    {
                        Vis* v=(Vis*)xalloc(sizeof(Vis));
                        v->orig_visual=pVisual;
			v->vid=pVisual->vid;
                        v->usegl=GL_TRUE;
                        v->rgb=GL_FALSE;
                        v->db=db;
                        v->alpha=GL_FALSE;
                        v->stencil=stencil;
                        v->depthb=depth;
                        v->accum=GL_FALSE;
                        v->next=0;
                        if (vtail)
                            vtail->next=v;
                        else
                            vhead=v;
                        vtail=v;
                        gl_visuals++;
                    }
                }
            }
        }
    }

    ErrorF("nvisual=%d (%d,%d)\n", nvisual, nongl_visuals, gl_visuals);
    /* If we are making synthetic visuals, we need to give them pointers
     * to  the appropriate place in the server "depths" list.
     * But since we are using the ORIGNAL visuals, there is no need
     */
    
    /*Then, would alloc space for fake VisualRec structs.
     * But again, since we dont use synthetic visuals, we dont
     * need that. (Right?)
     */


    /* If __vislist has data, add to end. Otherwise,
     * set our list to be the head
     */
    v=__vislist;
    if (v)
    {
        while (v->next)
            v=v->next;

        v->next=vhead;
    }
    else
    {
        __vislist=vhead;
    }
    return TRUE;
}

Bool GlxInitVisuals (VisualPtr   *visualp, 
		     DepthPtr    *depthp,
		     int         *nvisualp, 
		     int         *ndepthp,
		     int         *rootDepthp,
		     VisualID    *defaultVisp,
		     unsigned long   sizes,
		     int     bitsPerRGB )
{
	fprintf(stderr,"DEBUG: GlxInitVisuals called\n");

    {  
        extern GLboolean (*hwInitVisuals)(VisualPtr *, DepthPtr *, int *, int *, int *, VisualID *, unsigned long, int);          
        if (hwInitVisuals) return ((*hwInitVisuals)(visualp, depthp, nvisualp, ndepthp, rootDepthp, defaultVisp, sizes, bitsPerRGB
));   
    } 
    return (XSMesaInitVisuals(visualp, depthp, nvisualp, ndepthp, rootDepthp, defaultVisp, sizes, bitsPerRGB));                   
}

/* DEBUG routine */
static void dumpvis(){
	int vcount=0;
	Vis *v;
	for (v = __vislist; v!=0; v=v->next){
		fprintf(stderr,"  DEBUG: __vislist[%d] is %ld, (orig %ld)\n",
			vcount,v->vid,v->orig_visual->vid);

		vcount++;
	}
}

/* REWRITTEN from original utah-glx stuff.
 * XSMesaInitVisual() has also been drastically rewritten.
 * What we do now:
 * - Look at each entry in __vislist
 * -  Allocates __glScreens[screennum].visuals[], then calls
 * - Creates a new GLX-specific 'visual' entry in the above,
 *    to match each entry in the  tmp __vislist list,
 *    using XSMesaCreateVisual()
 *
 * Right now, I HARDCODE all values to XSMesaCreateVisual, ignoring
 *  everything but  __vislist->orig_visual
 *
 * XXX Furthermore, this is hardcoded to only deal with screen 0.
 *     I'm in a hurry, all this hacky garbage is getting to me.
 *
 *  FREES __vislist on exit
 */
void
GLFinalizeVisuals()
{
	int i;
	Vis* v;
        ScreenPtr pScreen=screenInfo.screens[0];
        GLScreen* glscreen=&__glScreens[0];
        int numVis=0;

	fprintf(stderr,"DEBUG: GLFinalizeVisuals() dumping vislist\n");
	dumpvis();

	v=__vislist; /* first entry */
	while(v!=0){
		numVis++;
		v=v->next;
	}
	v=__vislist;

	if(numVis==0){
		fprintf(stderr,"ERROR: no visuals for GLFinalizeVisuals\n");
		return;
	}

        glscreen->visuals=(XSMesaVisual*)xalloc(sizeof(XSMesaVisual)*numVis);
        glscreen->numVisuals=numVis;

	for(i=0;i<numVis;i++){
		int rgb=1, alpha=1, db=1, depthb=1, stencil=1,accum=1;
		int ximage_flag=1; /* 1 means use ximage, 0 means use pixmap*/

		fprintf(stderr,"DEBUG: GLFinalizeVisual creating wrapper for vis %ld\n",
				v->orig_visual->vid);

		/* This actually calls XSMesaCreateVisual() */
                glscreen->visuals[i]=
#ifdef NOTNOW
			GLXProcs.CreateVisual(	pScreen,
						v->orig_visual,
						v->rgb, v->alpha,
						v->db,
						ximage_flag,
						(v->depthb)*DEPTH_BITS,
						(v->stencil)*STENCIL_BITS,
						(v->accum)*ACCUM_BITS,
						0 /* overlay/underlay */
						);
#else
			GLXProcs.CreateVisual(	pScreen,
						v->orig_visual,
						rgb,alpha,db,
						ximage_flag,
						depthb*DEPTH_BITS,
						stencil*STENCIL_BITS,
						accum*ACCUM_BITS,
						0 /* overlay/underlay */
						);
#endif
		v=v->next;
	}

	v=__vislist;
	while (v)
	{
		Vis* next=v->next;
		xfree(v);
		v=next;
	}
	__vislist=0;
}

XSMesaVisual GLfind_vis(VisualID vid)
{
    int i;
    int screen;
    for (screen=0;screen<screenInfo.numScreens;screen++)
    {
        ScreenPtr pScreen=screenInfo.screens[screen];
        int numVis=pScreen->numVisuals;
        GLScreen* glscreen=&__glScreens[screen];
        for (i=0;i<numVis;i++)
        {
            if (glscreen->visuals[i]->pVisual->vid == vid)
                return glscreen->visuals[i];
        }
    }
    return 0;
}

